from canvas_sdk.commands.base import _BaseCommand as BaseCommand
from canvas_sdk.commands.commands.goal import GoalCommand


class CloseGoalCommand(BaseCommand):
    """A class for managing a CloseGoal command within a specific note."""

    class Meta:
        key = "closeGoal"

    goal_id: int | None = None
    achievement_status: GoalCommand.AchievementStatus | None = None
    progress: str | None = None


__exports__ = ("CloseGoalCommand",)



from typing import Any

from canvas_sdk.effects.base import EffectType, _BaseEffect


class CreateQuestionnaireResult(_BaseEffect):
    """CreateQuestionnaireResult effect."""

    class Meta:
        effect_type = EffectType.CREATE_QUESTIONNAIRE_RESULT

    interview_id: str | None = None
    score: float | None = None
    abnormal: bool | None = None
    narrative: str | None = None
    code_system: str | None = None
    code: str | None = None

    @property
    def values(self) -> dict[str, Any]:
        """Make the payload."""
        return {
            "interview_id": self.interview_id,
            "score": self.score,
            "abnormal": self.abnormal or False,
            "narrative": self.narrative or "",
            "code_system": self.code_system,
            "code": self.code,
        }


__exports__ = ("CreateQuestionnaireResult",)



from django.db import models


class DetectedIssue(models.Model):
    """DetectedIssue."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_detectedissue_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    identified = models.DateTimeField()
    deleted = models.BooleanField()
    originator = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    committer = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    entered_in_error = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="detected_issues", null=True
    )
    code = models.CharField()
    status = models.CharField()
    severity = models.CharField()
    reference = models.CharField()
    issue_identifier = models.CharField()
    issue_identifier_system = models.CharField()
    detail = models.TextField()


class DetectedIssueEvidence(models.Model):
    """DetectedIssueEvidence."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_detectedissueevidence_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    system = models.CharField()
    version = models.CharField()
    code = models.CharField()
    display = models.CharField()
    user_selected = models.BooleanField()
    detected_issue = models.ForeignKey(
        DetectedIssue,
        on_delete=models.DO_NOTHING,
        related_name="evidence",
        null=True,
    )


__exports__ = ("DetectedIssue", "DetectedIssueEvidence")



from canvas_sdk.effects.billing_line_item.add_billing_line_item import (
    AddBillingLineItem,
)
from canvas_sdk.effects.billing_line_item.remove_billing_line_item import (
    RemoveBillingLineItem,
)
from canvas_sdk.effects.billing_line_item.update_billing_line_item import (
    UpdateBillingLineItem,
)

__all__ = __exports__ = (
    "AddBillingLineItem",
    "UpdateBillingLineItem",
    "RemoveBillingLineItem",
)



from decimal import Decimal
from enum import Enum

from pydantic import Field, conlist

from canvas_sdk.commands.base import _BaseCommand
from canvas_sdk.commands.constants import ClinicalQuantity


class PrescribeCommand(_BaseCommand):
    """A class for managing a Prescribe command within a specific note."""

    class Meta:
        key = "prescribe"

    class Substitutions(Enum):
        ALLOWED = "allowed"
        NOT_ALLOWED = "not_allowed"

    fdb_code: str | None = Field(default=None, json_schema_extra={"commands_api_name": "prescribe"})
    icd10_codes: conlist(str, max_length=2) = Field(  # type: ignore[valid-type]
        default=[], json_schema_extra={"commands_api_name": "indications"}
    )
    sig: str = ""
    days_supply: int | None = None
    quantity_to_dispense: Decimal | float | int | None = None
    type_to_dispense: ClinicalQuantity | None = None
    refills: int | None = None
    substitutions: Substitutions | None = None
    pharmacy: str | None = None
    prescriber_id: str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "prescriber"}
    )
    supervising_provider_id: str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "supervising_provider"}
    )
    note_to_pharmacist: str | None = None

    @property
    def values(self) -> dict:
        """The Prescribe command's field values."""
        values = super().values

        if self.is_dirty("quantity_to_dispense"):
            values["quantity_to_dispense"] = (
                str(Decimal(self.quantity_to_dispense)) if self.quantity_to_dispense else None
            )

        return values


__exports__ = (
    "PrescribeCommand",
    # Not defined here but used in a current plugin
    "ClinicalQuantity",
    "Decimal",
)



import datetime
import json
from abc import ABC
from dataclasses import dataclass
from typing import Any
from uuid import UUID

from pydantic_core import InitErrorDetails

from canvas_generated.messages.effects_pb2 import Effect
from canvas_sdk.base import TrackableFieldsModel
from canvas_sdk.v1.data import PracticeLocation, Staff
from canvas_sdk.v1.data.appointment import AppointmentProgressStatus


@dataclass
class AppointmentIdentifier:
    """
    Dataclass for appointment identifiers.

    Attributes:
        system (str): The system identifier for the appointment.
        value (str): The value associated with the system identifier.
    """

    system: str
    value: str


class NoteOrAppointmentABC(TrackableFieldsModel, ABC):
    """
    Base class for all note effects.

    Attributes:
        practice_location_id (UUID | str): The ID of the practice location.
        provider_id (str): The ID of the provider.
    """

    class Meta:
        effect_type = None

    practice_location_id: UUID | str
    provider_id: str

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        """
        Validates the practice location and provider IDs and returns a list of error details if validation fails.

        Args:
            method (Any): The method being validated.

        Returns:
            list[InitErrorDetails]: A list of error details for invalid practice location or provider IDs.
        """
        errors = super()._get_error_details(method)

        if not PracticeLocation.objects.filter(id=self.practice_location_id).exists():
            errors.append(
                self._create_error_detail(
                    "value",
                    f"Practice location with ID {self.practice_location_id} does not exist.",
                    self.practice_location_id,
                )
            )

        if not Staff.objects.filter(id=self.provider_id).exists():
            errors.append(
                self._create_error_detail(
                    "value",
                    "Provider with ID {self.provider_id} does not exist.",
                    self.provider_id,
                )
            )

        return errors

    def create(self) -> Effect:
        """Originate a new command in the note body."""
        self._validate_before_effect("create")
        return Effect(
            type=f"CREATE_{self.Meta.effect_type}",
            payload=json.dumps(
                {
                    "data": self.values,
                }
            ),
        )


class AppointmentABC(NoteOrAppointmentABC, ABC):
    """
    Base class for appointment creation effects.

    Attributes:
        start_time (datetime.datetime): The start time of the appointment.
        duration_minutes (int): The duration of the appointment in minutes.
        status (AppointmentProgressStatus | None): The status of the appointment.
        external_identifiers (list[AppointmentIdentifier] | None): List of external identifiers for the appointment.
    """

    class Meta:
        effect_type = "APPOINTMENT"

    _dirty_excluded_keys = [
        "external_identifiers",
    ]

    start_time: datetime.datetime
    duration_minutes: int
    status: AppointmentProgressStatus | None = None
    external_identifiers: list[AppointmentIdentifier] | None = None

    @property
    def values(self) -> dict:
        """
        Returns a dictionary of modified attributes with type-specific transformations.
        """
        values = super().values

        if self.external_identifiers:
            values["external_identifiers"] = [
                {"system": identifier.system, "value": identifier.value}
                for identifier in self.external_identifiers
            ]

        return values


__exports__ = ("AppointmentIdentifier", "NoteOrAppointmentABC", "AppointmentABC")



from django.db import models


class Device(models.Model):
    """Device."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_device_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    originator = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    committer = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    entered_in_error = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="devices", null=True
    )
    note_id = models.BigIntegerField()
    deleted = models.BooleanField()
    labeled_contains_NRL = models.BooleanField()
    assigning_authority = models.CharField()
    scoping_entity = models.CharField()
    udi = models.CharField()
    di = models.CharField()
    issuing_agency = models.CharField()
    lot_number = models.CharField()
    brand_name = models.CharField()
    mri_safety_status = models.CharField()
    version_model_number = models.CharField()
    company_name = models.CharField()
    gmdnPTName = models.TextField()
    status = models.CharField()
    expiration_date = models.DateField()
    expiration_date_original = models.CharField()
    serial_number = models.CharField()
    manufacturing_date_original = models.CharField()
    manufacturing_date = models.DateField()
    manufacturer = models.CharField()
    procedure_id = models.BigIntegerField()


__exports__ = ("Device",)



from collections.abc import Container
from typing import cast

from django.db import models
from django.db.models import Q

from canvas_sdk.v1.data.base import (
    BaseModelManager,
    BaseQuerySet,
    CommittableQuerySetMixin,
    ForPatientQuerySetMixin,
    ValueSetLookupByNameQuerySet,
)


class ResponseOptionSet(models.Model):
    """ResponseOptionSet."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_responseoptionset_001"

    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    status = models.CharField()
    name = models.CharField()
    code_system = models.CharField()
    code = models.CharField()
    type = models.CharField()
    use_in_shx = models.BooleanField()


class ResponseOption(models.Model):
    """ResponseOption."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_responseoption_001"

    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    status = models.CharField()
    name = models.CharField()
    code = models.CharField()
    code_description = models.CharField()
    value = models.CharField()
    response_option_set = models.ForeignKey(
        ResponseOptionSet, on_delete=models.DO_NOTHING, related_name="options", null=True
    )
    ordering = models.IntegerField()


class Question(models.Model):
    """Question."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_question_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    status = models.CharField()
    name = models.CharField()
    response_option_set = models.ForeignKey(
        ResponseOptionSet, on_delete=models.DO_NOTHING, related_name="questions", null=True
    )
    acknowledge_only = models.BooleanField()
    show_prologue = models.BooleanField()
    code_system = models.CharField()
    code = models.CharField()


class QuestionnaireValueSetLookupQuerySet(ValueSetLookupByNameQuerySet):
    """QuerySet class for Questionaire ValueSet lookups."""

    @staticmethod
    def q_object(system: str, codes: Container[str]) -> Q:
        """The code system and code values for a Questionnaire are just attributes on the model."""
        return Q(code_system=system, code__in=codes)


class Questionnaire(models.Model):
    """Questionnaire."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_questionnaire_001"

    objects = models.Manager.from_queryset(QuestionnaireValueSetLookupQuerySet)()

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    status = models.CharField()
    name = models.CharField()
    expected_completion_time = models.FloatField()
    can_originate_in_charting = models.BooleanField()
    use_case_in_charting = models.CharField()
    scoring_function_name = models.TextField()
    scoring_code_system = models.CharField()
    scoring_code = models.CharField()
    code_system = models.CharField()
    code = models.CharField()
    search_tags = models.CharField()
    questions = models.ManyToManyField(Question, through="v1.QuestionnaireQuestionMap")
    use_in_shx = models.BooleanField()
    carry_forward = models.TextField()


class QuestionnaireQuestionMap(models.Model):
    """QuestionnaireQuestionMap."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_questionnairequestionmap_001"

    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    status = models.CharField()
    questionnaire = models.ForeignKey(Questionnaire, on_delete=models.DO_NOTHING, null=True)
    question = models.ForeignKey(Question, on_delete=models.DO_NOTHING, null=True)


class InterviewQuerySet(BaseQuerySet, ForPatientQuerySetMixin, CommittableQuerySetMixin):
    """InterviewQuerySet."""

    pass


InterviewManager = BaseModelManager.from_queryset(InterviewQuerySet)


class Interview(models.Model):
    """Interview."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_interview_001"

    objects = cast(InterviewQuerySet, InterviewManager())

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    deleted = models.BooleanField()
    committer = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    entered_in_error = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    status = models.CharField()
    name = models.CharField()
    language_id = models.BigIntegerField()
    use_case_in_charting = models.CharField()
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="interviews", null=True
    )
    note_id = models.BigIntegerField()
    appointment_id = models.BigIntegerField()
    questionnaires = models.ManyToManyField(
        Questionnaire,
        through="v1.InterviewQuestionnaireMap",
    )
    progress_status = models.CharField()
    created = models.DateTimeField()
    modified = models.DateTimeField()


class InterviewQuestionnaireMap(models.Model):
    """InterviewQuestionnaireMap."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_interviewquestionnairemap_001"

    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    status = models.CharField()
    interview = models.ForeignKey(Interview, on_delete=models.DO_NOTHING, null=True)
    questionnaire = models.ForeignKey(Questionnaire, on_delete=models.DO_NOTHING, null=True)


class InterviewQuestionResponse(models.Model):
    """InterviewQuestionResponse."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_interviewquestionresponse_001"

    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    status = models.CharField()
    interview = models.ForeignKey(
        Interview, on_delete=models.DO_NOTHING, related_name="interview_responses", null=True
    )
    questionnaire = models.ForeignKey(
        Questionnaire, on_delete=models.DO_NOTHING, related_name="interview_responses", null=True
    )
    question = models.ForeignKey(
        Question, on_delete=models.DO_NOTHING, related_name="interview_responses", null=True
    )
    response_option = models.ForeignKey(
        ResponseOption, on_delete=models.DO_NOTHING, related_name="interview_responses", null=True
    )
    response_option_value = models.TextField()
    questionnaire_state = models.TextField()
    interview_state = models.TextField()
    comment = models.CharField()


__exports__ = (
    "ResponseOptionSet",
    "ResponseOption",
    "Question",
    "Questionnaire",
    "QuestionnaireQuestionMap",
    "InterviewQuerySet",
    "Interview",
    "InterviewQuestionnaireMap",
    "InterviewQuestionResponse",
)



import json
from typing import Any
from uuid import UUID

from pydantic_core import InitErrorDetails

from canvas_generated.messages.effects_pb2 import Effect
from canvas_sdk.base import TrackableFieldsModel
from canvas_sdk.v1.data import Message as MessageModel
from canvas_sdk.v1.data import Patient, Staff


class Message(TrackableFieldsModel):
    """
    Effect to create and/or send a message.
    """

    class Meta:
        effect_type = "MESSAGE"

    message_id: str | UUID | None = None
    content: str
    sender_id: str | UUID
    recipient_id: str | UUID

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        if (
            not Patient.objects.filter(id=self.sender_id).exists()
            and not Staff.objects.filter(id=self.sender_id).exists()
        ):
            errors.append(
                self._create_error_detail(
                    "value",
                    f"Sender with ID {self.sender_id} does not exist.",
                    self.sender_id,
                )
            )

        if (
            not Patient.objects.filter(id=self.recipient_id).exists()
            and not Staff.objects.filter(id=self.recipient_id).exists()
        ):
            errors.append(
                self._create_error_detail(
                    "value",
                    f"Recipient with ID {self.recipient_id} does not exist.",
                    self.recipient_id,
                )
            )

        if method == "edit":
            if not self.message_id:
                errors.append(
                    self._create_error_detail(
                        "value",
                        "Message ID is required when editing a message.",
                        self.message_id,
                    )
                )
            else:
                if not MessageModel.objects.filter(id=self.message_id).exists():
                    errors.append(
                        self._create_error_detail(
                            "value",
                            f"Can't edit message with ID {self.message_id}: Does not exist.",
                            self.message_id,
                        )
                    )
        elif (
            method
            in (
                "create",
                "create_and_send",
            )
            and self.message_id
        ):
            errors.append(
                self._create_error_detail(
                    "value",
                    "Can't set message ID when creating a message.",
                    self.message_id,
                )
            )

        return errors

    def create(self) -> Effect:
        """Originate a new command in the note body."""
        self._validate_before_effect("create")
        return Effect(
            type=f"CREATE_{self.Meta.effect_type}",
            payload=json.dumps(
                {
                    "data": self.values,
                }
            ),
        )

    def create_and_send(self) -> Effect:
        """Create and send message."""
        self._validate_before_effect("create_and_send")
        return Effect(
            type="CREATE_AND_SEND_MESSAGE",
            payload=json.dumps({"data": self.values}),
        )

    def edit(self) -> Effect:
        """Edit message."""
        self._validate_before_effect("edit")
        return Effect(type="EDIT_MESSAGE", payload=json.dumps({"data": self.values}))

    def send(self) -> Effect:
        """Send message."""
        self._validate_before_effect("send")
        return Effect(
            type="SEND_MESSAGE",
            payload=json.dumps({"data": self.values}),
        )


__exports__ = ("Message",)



import ast
import json
import tarfile
import tempfile
from collections.abc import Iterable
from pathlib import Path
from typing import Any, cast
from urllib.parse import urljoin

import requests
import typer
from cookiecutter.exceptions import OutputDirExistsException
from cookiecutter.main import cookiecutter

from canvas_cli.apps.auth.utils import get_default_host, get_or_request_api_token
from canvas_cli.utils.context import context
from canvas_cli.utils.validators import validate_manifest_file


def plugin_url(host: str, *paths: str) -> str:
    """Generates the plugin url for managing plugins in a Canvas instance."""
    join = "/".join(["plugin-io/plugins", "/".join(paths or [])])
    join = join if join.endswith("/") else join + "/"

    return urljoin(host, join)


def validate_package(package: Path) -> Path:
    """Validate if `package` Path exists and it is a file."""
    if not package.exists():
        raise typer.BadParameter(f"Package {package} does not exist")

    if not package.is_file():
        raise typer.BadParameter(f"Package {package} isn't a file")

    if not package.name.endswith("tar.gz") and not package.name.endswith("whl"):
        raise typer.BadParameter(f"Package {package} needs to be a tar.gz or a whl")

    return package


def _build_package(package: Path) -> Path:
    """Compresses `package` and returns the built archive, ignoring symlinks, hidden folders, and hidden files."""
    package = package.resolve()

    if not package.exists() or not package.is_dir():
        raise typer.BadParameter(f"Couldn't build {package}, not a dir")

    with tempfile.NamedTemporaryFile(suffix=".tar.gz", delete=False) as file:
        with tarfile.open(file.name, "w:gz") as tar:
            for root in package.rglob("*"):
                # Skip hidden files and directories (starting with '.') and symlinks
                if any(part.startswith(".") for part in root.parts) or root.is_symlink():
                    continue

                tar.add(root, arcname=root.relative_to(package))
        return Path(file.name)


def _get_name_from_metadata(host: str, token: str, package: Path) -> str | None:
    """Extract metadata from a provided package and return the package name if it exists in the metadata."""
    try:
        with open(package, "rb") as package_file:
            metadata_response = requests.post(
                plugin_url(host, "extract-metadata"),
                headers={"Authorization": f"Bearer {token}"},
                files={"package": package_file},
            )
    except requests.exceptions.RequestException:
        print(f"Failed to connect to {host}")
        raise typer.Exit(1) from None

    if metadata_response.status_code != requests.codes.ok:
        print(f"Status code {metadata_response.status_code}: {metadata_response.text}")
        raise typer.Exit(1)

    metadata = metadata_response.json()
    return metadata.get("name")


def _get_meta_class(text: str, classname: str) -> ast.ClassDef | None:
    tree = ast.parse(text)
    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef) and node.name == classname:
            class_def = node
            for b in class_def.body:
                if isinstance(b, ast.ClassDef) and b.name == "Meta":
                    return b
    return None


def _get_meta_properties(protocol_path: Path, classname: str) -> dict[str, str]:
    meta: dict[str, str] = {}

    if not protocol_path.exists():
        return meta

    meta_class = _get_meta_class(protocol_path.read_text(), classname)
    if not meta_class:
        return meta

    for meta_b in meta_class.body:
        if not isinstance(meta_b, ast.Assign | ast.AnnAssign):
            continue
        targets = [meta_b.target] if isinstance(meta_b, ast.AnnAssign) else meta_b.targets
        target_id = next((t.id for t in targets if isinstance(t, ast.Name)), None)
        if not target_id:
            continue
        if isinstance(meta_b.value, ast.Constant):
            value = meta_b.value.value
        elif isinstance(meta_b.value, ast.List):
            value = [cast(ast.Constant, e).value for e in meta_b.value.elts]
        elif isinstance(meta_b.value, ast.Dict):
            keys = meta_b.value.keys
            values = meta_b.value.values
            value = {
                cast(ast.Constant, k).value: cast(ast.Constant, values[i]).value
                for i, k in enumerate(keys)
            }
        else:
            value = None
        meta[target_id] = value

    return meta


def _get_protocols_with_new_cqm_properties(
    protocol_classes: Iterable[dict[str, Any]], plugin: Path
) -> Iterable[dict[str, Any]] | None:
    """Extract the meta properties of any ClinicalQualityMeasure Protocols included in the plugin if they have changed."""
    has_updates = False
    protocol_props = []
    for p in protocol_classes:
        mod, classname = p["class"].split(":")
        mod = mod.replace(f"{plugin.name}.", "").replace(".", "/") + ".py"
        p_path: Path = plugin / mod
        meta = _get_meta_properties(p_path, classname)
        if meta and meta != p.get("meta"):
            has_updates = True
            protocol_props.append(p | {"meta": meta})
        else:
            protocol_props.append(p)

    return protocol_props if has_updates else None


def get_base_plugin_template_path(plugin_type: str) -> Path:
    """Return context's base_plugin_template_path, so it can be used as a Typer default."""
    match plugin_type:
        case "application":
            return context.plugin_template_dir / "application"
        case _:
            return context.plugin_template_dir / context.default_plugin_template_name


def init(
    plugin_type: str = typer.Argument(
        "protocol",
        help="The type of plugin to create. Options are 'application' or 'protocol'.",
    ),
) -> None:
    """Create a new plugin."""
    template = get_base_plugin_template_path(plugin_type)
    try:
        project_dir = cookiecutter(str(template))
    except OutputDirExistsException:
        raise typer.BadParameter("The supplied directory already exists") from None

    print(f"Project created in {project_dir}")


def install(
    plugin_name: Path = typer.Argument(..., help="Path to plugin to install"),
    host: str | None = typer.Option(
        callback=get_default_host,
        help="Canvas instance to connect to",
        default=None,
    ),
) -> None:
    """Install a plugin into a Canvas instance."""
    if not host:
        raise typer.BadParameter("Please specify a host or add one to the configuration file")

    token = get_or_request_api_token(host)

    if not plugin_name.exists():
        raise typer.BadParameter(f"Plugin '{plugin_name}' does not exist")

    if plugin_name.is_dir():
        validate_manifest(plugin_name)
        built_package_path = _build_package(plugin_name)
    else:
        raise typer.BadParameter(f"Plugin '{plugin_name}' needs to be a valid directory")

    print(f"Installing plugin: {built_package_path} into {host}")

    url = plugin_url(host)

    print(f"Posting {built_package_path.absolute()} to {url}")

    try:
        with open(built_package_path, "rb") as package:
            r = requests.post(
                url,
                data={"is_enabled": True},
                files={"package": package},
                headers={"Authorization": f"Bearer {token}"},
            )
    except requests.exceptions.RequestException:
        print(f"Failed to connect to {host}")
        raise typer.Exit(1) from None

    if r.status_code == requests.codes.created:
        print(f"Plugin {plugin_name} successfully installed!")

    # If we got a conflict, means there's a duplicate plugin and install can't handle that.
    # So we need to get the plugin-name from the package and call `update` directly
    elif r.status_code == requests.codes.conflict and (
        package_name := _get_name_from_metadata(host, token, built_package_path)
    ):
        print(f"Plugin {package_name} already exists, updating instead...")
        update(package_name, built_package_path, is_enabled=True, host=host)
    else:
        print(f"Status code {r.status_code}: {r.text}")
        raise typer.Exit(1)


def uninstall(
    name: str = typer.Argument(..., help="Plugin name to uninstall"),
    force: bool = typer.Option(
        False, "--force", help="Force uninstallation of the plugin", show_default=False
    ),
    host: str | None = typer.Option(
        callback=get_default_host,
        help="Canvas instance to connect to",
        default=None,
    ),
) -> None:
    """Uninstall a plugin from a Canvas instance."""
    if not host:
        raise typer.BadParameter("Please specify a host or or add one to the configuration file")

    url = plugin_url(host, name)

    print(f"Uninstalling {name} using {url}")

    token = get_or_request_api_token(host)

    try:
        r = requests.delete(
            url,
            headers={
                "Authorization": f"Bearer {token}",
            },
            params={"force": str(force)},
        )
    except requests.exceptions.RequestException:
        print(f"Failed to connect to {host}")
        raise typer.Exit(1) from None

    if r.status_code == requests.codes.no_content:
        print(f"Plugin {name} successfully uninstalled!")
    else:
        print(f"Status code {r.status_code}: {r.text}")
        raise typer.Exit(1)


def enable(
    name: str = typer.Argument(..., help="Plugin name to enable"),
    host: str | None = typer.Option(
        callback=get_default_host,
        help="Canvas instance to connect to",
        default=None,
    ),
) -> None:
    """Enable a plugin from a Canvas instance."""
    if not host:
        raise typer.BadParameter("Please specify a host or or add one to the configuration file")

    url = plugin_url(host, name)

    print(f"Enabling {name} using {url}")

    token = get_or_request_api_token(host)

    try:
        r = requests.patch(
            url,
            data={"is_enabled": True},
            headers={
                "Authorization": f"Bearer {token}",
            },
        )
    except requests.exceptions.RequestException:
        print(f"Failed to connect to {host}")
        raise typer.Exit(1) from None

    if r.ok:
        print(f"Plugin {name} successfully enabled!")
    else:
        print(f"Status code {r.status_code}: {r.text}")
        raise typer.Exit(1)


def disable(
    name: str = typer.Argument(..., help="Plugin name to disable"),
    host: str | None = typer.Option(
        callback=get_default_host,
        help="Canvas instance to connect to",
        default=None,
    ),
) -> None:
    """Disable a plugin from a Canvas instance."""
    if not host:
        raise typer.BadParameter("Please specify a host or or add one to the configuration file")

    url = plugin_url(host, name)

    print(f"Disabling {name} using {url}")

    token = get_or_request_api_token(host)

    try:
        r = requests.patch(
            url,
            data={"is_enabled": False},
            headers={
                "Authorization": f"Bearer {token}",
            },
        )
    except requests.exceptions.RequestException:
        print(f"Failed to connect to {host}")
        raise typer.Exit(1) from None

    if r.ok:
        print(f"Plugin {name} successfully disabled!")
    else:
        print(f"Status code {r.status_code}: {r.text}")
        raise typer.Exit(1)


def list(
    host: str | None = typer.Option(
        callback=get_default_host,
        help="Canvas instance to connect to",
        default=None,
    ),
) -> None:
    """List all plugins from a Canvas instance."""
    if not host:
        raise typer.BadParameter("Please specify a host or add one to the configuration file")

    url = plugin_url(host)

    token = get_or_request_api_token(host)

    try:
        r = requests.get(
            url,
            headers={"Authorization": f"Bearer {token}"},
        )
    except requests.exceptions.RequestException:
        print(f"Failed to connect to {host}")
        raise typer.Exit(1) from None

    if r.status_code == requests.codes.ok:
        plugins = r.json().get("results", [])
        if not plugins:
            print(f"No plugins are currently installed on {host}")
        for plugin in plugins:
            print(
                f"{plugin['name']}@{plugin['version']}\t{'enabled' if plugin['is_enabled'] else 'disabled'}"
            )
    else:
        print(f"Status code {r.status_code}: {r.text}")
        raise typer.Exit(1)


def validate_manifest(
    plugin_name: Path = typer.Argument(..., help="Path to plugin to validate"),
) -> None:
    """Validate the Canvas Manifest json file."""
    if not plugin_name.exists():
        raise typer.BadParameter(f"Plugin {plugin_name} does not exist")

    if not plugin_name.is_dir():
        raise typer.BadParameter(f"Plugin {plugin_name} is not a directory, nothing to validate")

    manifest = plugin_name / "CANVAS_MANIFEST.json"

    if not manifest.exists():
        raise typer.BadParameter(
            f"Plugin {plugin_name} does not have a CANVAS_MANIFEST.json file to validate"
        )

    try:
        manifest_json = json.loads(manifest.read_text())
        protocols = manifest_json.get("components", {}).get("protocols", [])
        if new_protocols := _get_protocols_with_new_cqm_properties(protocols, plugin_name):
            print(
                f"Updating the CANVAS_MANIFEST.json file for {plugin_name} with CQM meta properties"
            )
            manifest_json["components"]["protocols"] = new_protocols
            manifest.write_text(json.dumps(manifest_json))
            manifest_json = json.loads(manifest.read_text())

    except json.JSONDecodeError:
        print("There was a problem loading the manifest file, please ensure it's valid JSON")
        raise typer.Abort() from None

    validate_manifest_file(manifest_json)

    print(f"Plugin {plugin_name} has a valid CANVAS_MANIFEST.json file")


def update(
    name: str = typer.Argument(..., help="Plugin name to update"),
    package_path: Path | None = typer.Option(
        help="Path to a wheel or sdist file containing the python package to install",
        default=None,
    ),
    is_enabled: bool | None = typer.Option(
        None, "--enable/--disable", show_default=False, help="Enable/disable the plugin"
    ),
    host: str | None = typer.Option(
        callback=get_default_host,
        help="Canvas instance to connect to",
        default=None,
    ),
) -> None:
    """Updates a plugin from an instance."""
    if not host:
        raise typer.BadParameter("Please specify a host or set a default via the `auth` command")

    if package_path:
        validate_package(package_path)

    token = get_or_request_api_token(host)

    print(f"Updating plugin {name} from {host} with {is_enabled=}, {package_path=}")

    url = plugin_url(host, name)

    try:
        data = {"is_enabled": is_enabled} if is_enabled is not None else {}
        headers = {"Authorization": f"Bearer {token}"}

        if package_path:
            with open(package_path, "rb") as package:
                r = requests.patch(
                    url,
                    data=data,
                    headers=headers,
                    files={"package": package},
                )
        else:
            r = requests.patch(
                url,
                data=data,
                headers=headers,
            )
    except requests.exceptions.RequestException:
        print(f"Failed to connect to {host}")
        raise typer.Exit(1) from None

    if r.status_code == requests.codes.ok:
        print("Plugin successfully updated!")

    else:
        print(f"Status code {r.status_code}: {r.text}")
        raise typer.Exit(1)



from canvas_generated.messages.events_pb2 import Event as EventRequest
from canvas_generated.messages.events_pb2 import EventResponse, EventType

from .base import Event

__all__ = __exports__ = (
    "EventRequest",
    "EventResponse",
    "EventType",
    "Event",
)



from enum import Enum

from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand as BaseCommand
from canvas_sdk.commands.constants import ServiceProvider


class ImagingOrderCommand(BaseCommand):
    """A class for managing an Imaging Order command within a specific note."""

    class Meta:
        key = "imagingOrder"

    class Priority(Enum):
        ROUTINE = "Routine"
        URGENT = "Urgent"

    image_code: str | None = Field(default=None, json_schema_extra={"commands_api_name": "image"})
    diagnosis_codes: list[str] | None = Field(
        default=None, json_schema_extra={"commands_api_name": "indications"}
    )
    priority: Priority | None = None
    additional_details: str | None = None
    service_provider: ServiceProvider | None = Field(
        default=None, json_schema_extra={"commands_api_name": "ordering_provider"}
    )
    comment: str | None = None
    ordering_provider_key: str | None = None
    linked_items_urns: list[str] | None = None

    @property
    def values(self) -> dict:
        """The Imaging Order command's field values."""
        values = super().values

        if self.is_dirty("service_provider"):
            values["service_provider"] = self.service_provider.__dict__

        return values


__exports__ = ("ImagingOrderCommand",)



from django.contrib.postgres.fields import ArrayField
from django.db import models


class ReasonForVisitSettingCoding(models.Model):
    """ReasonForVisitSettingCoding."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_reasonforvisitsettingcoding_001"

    objects: models.Manager["ReasonForVisitSettingCoding"]

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)

    code = models.CharField()
    display = models.CharField()
    system = models.CharField()
    version = models.CharField()

    duration = ArrayField(models.DurationField())


__exports__ = ("ReasonForVisitSettingCoding",)



from typing import Any

from canvas_sdk.commands.constants import Coding
from canvas_sdk.effects.base import EffectType, _BaseEffect


class UpdateBillingLineItem(_BaseEffect):
    """
    An Effect that will update a billing line item in a note footer.
    """

    class Meta:
        effect_type = EffectType.UPDATE_BILLING_LINE_ITEM
        apply_required_fields = ("billing_line_item_id",)

    billing_line_item_id: str | None = None
    cpt: str | None = None
    units: int | None = None
    assessment_ids: list[str] | None = None
    modifiers: list[Coding] | None = None

    @property
    def values(self) -> dict[str, Any]:
        """The BillingLineItem's values."""
        values: dict[str, str | int | list] = {}
        # only include the values where a value has been set so as to not unintentionally override existing values
        if self.cpt is not None:
            values["cpt"] = self.cpt
        if self.units is not None:
            values["units"] = self.units
        if self.assessment_ids is not None:
            values["assessment_ids"] = self.assessment_ids
        if self.modifiers is not None:
            values["modifiers"] = self.modifiers

        return values

    @property
    def effect_payload(self) -> dict[str, Any]:
        """The payload of the effect."""
        return {
            "billing_line_item_id": self.billing_line_item_id,
            "data": self.values,
        }


__exports__ = ("UpdateBillingLineItem",)



from canvas_sdk.effects.note.appointment import Appointment, ScheduleEvent
from canvas_sdk.effects.note.base import AppointmentIdentifier
from canvas_sdk.effects.note.note import Note

__all__ = __exports__ = (
    "AppointmentIdentifier",
    "Note",
    "Appointment",
    "ScheduleEvent",
)



import datetime
import json
from dataclasses import dataclass
from typing import Any

from pydantic_core import InitErrorDetails

from canvas_generated.messages.effects_pb2 import Effect
from canvas_sdk.base import TrackableFieldsModel
from canvas_sdk.v1.data import PracticeLocation, Staff
from canvas_sdk.v1.data.common import ContactPointSystem, ContactPointUse, PersonSex


@dataclass
class PatientContactPoint:
    """A class representing a patient contact point."""

    system: ContactPointSystem
    value: str
    use: ContactPointUse
    rank: int
    has_consent: bool | None = None

    def to_dict(self) -> dict[str, Any]:
        """Convert the contact point to a dictionary."""
        return {
            "system": self.system.value,
            "value": self.value,
            "use": self.use.value,
            "rank": self.rank,
            "has_consent": self.has_consent,
        }


class Patient(TrackableFieldsModel):
    """Effect to create a Patient record."""

    class Meta:
        effect_type = "PATIENT"

    first_name: str
    last_name: str
    middle_name: str | None = None
    birthdate: datetime.date | None = None
    prefix: str | None = None
    suffix: str | None = None
    sex_at_birth: PersonSex | None = None
    nickname: str | None = None
    social_security_number: str | None = None
    administrative_note: str | None = None
    clinical_note: str | None = None
    default_location_id: str | None = None
    default_provider_id: str | None = None
    previous_names: list[str] | None = None
    contact_points: list[PatientContactPoint] | None = None

    @property
    def values(self) -> dict[str, Any]:
        """Return the values of the patient as a dictionary."""
        return {
            "first_name": self.first_name,
            "last_name": self.last_name,
            "middle_name": self.middle_name,
            "previous_names": self.previous_names or [],
            "birthdate": self.birthdate.isoformat() if self.birthdate else None,
            "prefix": self.prefix,
            "suffix": self.suffix,
            "sex_at_birth": self.sex_at_birth.value if self.sex_at_birth else None,
            "nickname": self.nickname,
            "social_security_number": self.social_security_number,
            "administrative_note": self.administrative_note,
            "clinical_note": self.clinical_note,
            "default_location": self.default_location_id,
            "default_provider": self.default_provider_id,
            "contact_points": [cp.to_dict() for cp in self.contact_points]
            if self.contact_points
            else None,
        }

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        if (
            self.default_location_id
            and not PracticeLocation.objects.filter(id=self.default_location_id).exists()
        ):
            errors.append(
                self._create_error_detail(
                    "value",
                    f"Practice location with ID {self.default_location_id} does not exist.",
                    self.default_location_id,
                )
            )

        if (
            self.default_provider_id
            and not Staff.objects.filter(id=self.default_provider_id).exists()
        ):
            errors.append(
                self._create_error_detail(
                    "value",
                    f"Provider with ID {self.default_provider_id} does not exist.",
                    self.default_provider_id,
                )
            )

        return errors

    def create(self) -> Effect:
        """Create a new Patient."""
        self._validate_before_effect("create")

        return Effect(
            type=f"CREATE_{self.Meta.effect_type}",
            payload=json.dumps(
                {
                    "data": self.values,
                }
            ),
        )


__exports__ = ("Patient", "PatientContactPoint")



from dataclasses import dataclass
from enum import StrEnum
from typing import Any

from pydantic_core import InitErrorDetails

from canvas_sdk.effects import EffectType, _BaseEffect


class InputType(StrEnum):
    """Type of input for a form field."""

    TEXT = "text"
    SELECT = "select"
    DATE = "date"


@dataclass
class FormField:
    """A class representing a Field."""

    key: str | None = None
    label: str | None = None
    required: bool = False
    editable: bool = True
    type: InputType = InputType.TEXT
    options: list[str] | None = None

    def to_dict(self) -> dict[str, Any]:
        """Convert the Field to a dictionary."""
        return {
            "key": self.key,
            "label": self.label,
            "required": self.required,
            "editable": self.editable,
            "type": self.type.value,
            "options": self.options,
        }


class PatientMetadataCreateFormEffect(_BaseEffect):
    """An Effect that will create a form."""

    class Meta:
        effect_type = EffectType.PATIENT_METADATA__CREATE_ADDITIONAL_FIELDS

    form_fields: list[FormField]

    @property
    def values(self) -> dict[str, Any]:
        """Return the values of the form as a dictionary."""
        return {"form": [field.to_dict() for field in self.form_fields]}

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        for field in self.form_fields:
            if field.type != InputType.SELECT and field.options:
                errors.append(
                    self._create_error_detail(
                        "value",
                        "The options attribute is only used for fields of type select",
                        field.key,
                    )
                )

        return errors


__exports__ = ("PatientMetadataCreateFormEffect", "FormField", "InputType")



import importlib.metadata
from abc import ABC, abstractmethod
from typing import Any

import deprecation

from canvas_sdk.effects import Effect
from canvas_sdk.events import Event

version = importlib.metadata.version("canvas")


class BaseHandler(ABC):
    """The class that all handlers inherit from."""

    environment: dict[str, Any]
    secrets: dict[str, Any]
    event: Event

    def __init__(
        self,
        event: Event,
        secrets: dict[str, Any] | None = None,
        environment: dict[str, Any] | None = None,
    ) -> None:
        self.event = event
        self.secrets = secrets or {}
        self.environment = environment or {}

    @property
    @deprecation.deprecated(
        deprecated_in="0.11.0",
        removed_in="1.0.0",
        current_version=version,
        details="Use 'event.context' directly instead",
    )
    def context(self) -> dict[str, Any]:
        """The context of the event."""
        return self.event.context

    @property
    @deprecation.deprecated(
        deprecated_in="0.11.0",
        removed_in="1.0.0",
        current_version=version,
        details="Use 'event.target.id' directly instead",
    )
    def target(self) -> str:
        """The target id of the event."""
        return self.event.target.id

    @abstractmethod
    def compute(self) -> list[Effect]:
        """Compute the effects to be applied."""
        pass

    def accept_event(self) -> bool:
        """Determine whether an event should be accepted and handled by the handler."""
        return True


__exports__ = ("BaseHandler",)



from enum import Enum
from uuid import UUID

from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand


class AssessCommand(_BaseCommand):
    """A class for managing an Assess command within a specific note."""

    class Meta:
        key = "assess"

    class Status(Enum):
        IMPROVED = "improved"
        STABLE = "stable"
        DETERIORATED = "deteriorated"

    condition_id: UUID | str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "condition"}
    )
    background: str | None = None
    status: Status | None = None
    narrative: str | None = None


# how do we make sure that condition_id is a valid condition for the patient?

# idea1:
# create a class attribute 'pre_validate_condition': bool
#      True: before doing any actions against the home-app instance, will first check
#            if its an active condition for that patient, and it not then logs a message
#            and doesn't do the action. can even create a cache so that we dont always
#            have to keep asking home-app for every patient
#      False: still tries to do it but will run up against whatever home-app barriers
#             we have against actions like that (if those dont already exist, then
#             definitely create them!)


# idea2:
# validator that checks that condition.patient is the same as note.patient

__exports__ = ("AssessCommand",)



import functools
import json
from collections.abc import Generator
from pathlib import Path
from typing import Any, TypedDict

import yaml
from jsonschema import Draft7Validator, validators

from canvas_sdk.utils.plugins import plugin_only


class Response(TypedDict):
    """A Response of a Questionnaire."""

    name: str
    code: str
    code_description: str
    value: str


class Question(TypedDict):
    """A Question of a Questionnaire."""

    code_system: str
    code: str
    code_description: str
    content: str
    responses_code_system: str
    responses_type: str
    display_result_in_social_history_section: bool
    responses: list[Response]


class QuestionnaireConfig(TypedDict):
    """A Questionnaire configuration."""

    name: str
    form_type: str
    code_system: str
    code: str
    can_originate_in_charting: bool
    prologue: str
    display_results_in_social_history_section: bool
    questions: list[Question]


def extend_with_defaults(validator_class: type[Draft7Validator]) -> type[Draft7Validator]:
    """Extend a Draft7Validator with default values for properties."""
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(
        validator: Draft7Validator,
        properties: dict[str, Any],
        instance: dict[str, Any],
        schema: dict[str, Any],
    ) -> Generator[Any, None, None]:
        for property, subschema in properties.items():
            if "default" in subschema:
                instance.setdefault(property, subschema["default"])

        yield from validate_properties(
            validator,
            properties,
            instance,
            schema,
        )

    return validators.extend(
        validator_class,
        {"properties": set_defaults},
    )


ExtendedDraft7Validator = extend_with_defaults(Draft7Validator)


@plugin_only
def from_yaml(questionnaire_name: str, **kwargs: Any) -> QuestionnaireConfig | None:
    """Load a Questionnaire configuration from a YAML file.

    Args:
        questionnaire_name (str): The path to the questionnaire file, relative to the plugin package.
            If the path starts with a forward slash ("/"), it will be stripped during resolution.
        kwargs (Any): Additional keyword arguments.

    Returns:
        QuestionnaireConfig: The loaded Questionnaire configuration.

    Raises:
        FileNotFoundError: If the questionnaire file does not exist within the plugin's directory
            or if the resolved path is invalid.
        PermissionError: If the resolved path is outside the plugin's directory.
        ValidationError: If the questionnaire file does not conform to the JSON schema.
    """
    plugin_dir = kwargs["plugin_dir"]
    questionnaire_config_path = Path(plugin_dir / questionnaire_name.lstrip("/")).resolve()

    if not questionnaire_config_path.is_relative_to(plugin_dir):
        raise PermissionError(f"Invalid Questionnaire '{questionnaire_name}'")
    elif not questionnaire_config_path.exists():
        raise FileNotFoundError(f"Questionnaire {questionnaire_name} not found.")

    questionnaire_config = yaml.load(questionnaire_config_path.read_text(), Loader=yaml.SafeLoader)
    ExtendedDraft7Validator(json_schema()).validate(questionnaire_config)

    return questionnaire_config


@functools.cache
def json_schema() -> dict[str, Any]:
    """Reads the JSON schema for a Questionnaire Config."""
    schema = json.loads(
        (Path(__file__).resolve().parent.parent.parent / "schemas/questionnaire.json").read_text()
    )

    return schema


__exports__ = (
    "Draft7Validator",
    "ExtendedDraft7Validator",
    "from_yaml",
)



import concurrent
import functools
import os
import urllib.parse
from collections.abc import Callable, Iterable, Mapping
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Literal, Protocol, TypeVar

import requests

from canvas_sdk.utils.metrics import measured

F = TypeVar("F", bound=Callable)


class _BatchableRequest:
    """Representation of a request that will be executed in parallel with other requests."""

    def __init__(
        self, method: Literal["GET", "POST", "PUT", "PATCH"], url: str, **kwargs: Any
    ) -> None:
        self._method = method
        self._url = url
        self._kwargs = kwargs

    def fn(self, client: "Http") -> Callable:
        """
        Return a callable constructed from an Http object and the method, URL, and kwargs.

        The callable is passed to the ThreadPoolExecutor.
        """
        client_method: Callable
        match self._method:
            case "GET":
                client_method = client.get
            case "POST":
                client_method = client.post
            case "PUT":
                client_method = client.put
            case "PATCH":
                client_method = client.patch
            case _:
                raise ValueError(f"HTTP method {self._method} is not supported")

        return functools.partial(client_method, self._url, **self._kwargs)


class BatchableRequest(Protocol):
    """Protocol for batchable requests."""

    def fn(self, client: "Http") -> Callable:
        """
        Return a callable that can be passed to the ThreadPoolExecutor.
        """
        ...


def batch_get(
    url: str, headers: Mapping[str, str | bytes | None] | None = None
) -> BatchableRequest:
    """Return a batchable GET request."""
    return _BatchableRequest("GET", url, headers=headers)


def batch_post(
    url: str,
    json: dict | None = None,
    data: dict | str | list | bytes | None = None,
    headers: Mapping[str, str | bytes | None] | None = None,
) -> BatchableRequest:
    """Return a batchable POST request."""
    return _BatchableRequest("POST", url, json=json, data=data, headeres=headers)


def batch_put(
    url: str,
    json: dict | None = None,
    data: dict | str | list | bytes | None = None,
    headers: Mapping[str, str | bytes | None] | None = None,
) -> BatchableRequest:
    """Return a batchable PUT request."""
    return _BatchableRequest("PUT", url, json=json, data=data, headers=headers)


def batch_patch(
    url: str,
    json: dict | None = None,
    data: dict | str | list | bytes | None = None,
    headers: Mapping[str, str | bytes | None] | None = None,
) -> BatchableRequest:
    """Return a batchable PATCH request."""
    return _BatchableRequest("PATCH", url, json=json, data=data, headers=headers)


class Http:
    """A helper class for completing HTTP calls with metrics tracking."""

    _MAX_REQUEST_TIMEOUT_SECONDS = 30

    _base_url: str
    _session: requests.Session

    def join_url(self, url: str) -> str:
        """
        Join a URL to the base_url.
        """
        joined = urllib.parse.urljoin(self._base_url, url)

        if not joined.startswith(self._base_url):
            raise ValueError("You may not access other URLs using this client.")

        return joined

    def __init__(self, base_url: str = "") -> None:
        self._base_url = base_url
        self._session = requests.Session()

    @measured(track_plugins_usage=True)
    def get(
        self, url: str, headers: Mapping[str, str | bytes | None] | None = None
    ) -> requests.Response:
        """Sends a GET request."""
        if headers is None:
            headers = {}
        return self._session.get(
            self.join_url(url),
            headers=headers,
            timeout=self._MAX_REQUEST_TIMEOUT_SECONDS,
        )

    @measured(track_plugins_usage=True)
    def post(
        self,
        url: str,
        json: dict | None = None,
        data: dict | str | list | bytes | None = None,
        headers: Mapping[str, str | bytes | None] | None = None,
    ) -> requests.Response:
        """Sends a POST request."""
        return self._session.post(
            self.join_url(url),
            json=json,
            data=data,
            headers=headers,
            timeout=self._MAX_REQUEST_TIMEOUT_SECONDS,
        )

    @measured(track_plugins_usage=True)
    def put(
        self,
        url: str,
        json: dict | None = None,
        data: dict | str | list | bytes | None = None,
        headers: Mapping[str, str | bytes | None] | None = None,
    ) -> requests.Response:
        """Sends a PUT request."""
        return self._session.put(
            self.join_url(url),
            json=json,
            data=data,
            headers=headers,
            timeout=self._MAX_REQUEST_TIMEOUT_SECONDS,
        )

    @measured(track_plugins_usage=True)
    def patch(
        self,
        url: str,
        json: dict | None = None,
        data: dict | str | list | bytes | None = None,
        headers: Mapping[str, str | bytes | None] | None = None,
    ) -> requests.Response:
        """Sends a PATCH request."""
        return self._session.patch(
            self.join_url(url),
            json=json,
            data=data,
            headers=headers,
            timeout=self._MAX_REQUEST_TIMEOUT_SECONDS,
        )

    @measured(track_plugins_usage=True)
    def batch_requests(
        self,
        batch_requests: Iterable[BatchableRequest],
        timeout: int | None = None,
    ) -> list[requests.Response]:
        """
        Execute requests in parallel.

        Wait for the responses to complete, and then return a list of the responses in the same
        ordering as the requests.
        """
        if timeout is None:
            timeout = self._MAX_REQUEST_TIMEOUT_SECONDS
        elif timeout < 1 or timeout > self._MAX_REQUEST_TIMEOUT_SECONDS:
            raise ValueError(
                "Timeout value must be greater than 0 and less than or equal "
                f"to {self._MAX_REQUEST_TIMEOUT_SECONDS} seconds"
            )

        with ThreadPoolExecutor() as executor:
            futures = [executor.submit(request.fn(self)) for request in batch_requests]

            concurrent.futures.wait(futures, timeout=timeout)

            return [future.result() for future in futures]


class JsonOnlyResponse:
    """
    A very simple Response analog that only allows access to the json() method
    and the status_code.

    If we returned the response directly the user could look at the request
    headers on the response object and see any authentication headers we sent
    on their behalf.
    """

    _json: dict[str, Any] | None
    status_code: int

    def __init__(self, response: requests.Response):
        self.status_code = response.status_code

        try:
            self._json = response.json()
        except Exception:
            self._json = None

    def json(self) -> dict[str, Any] | None:
        return self._json


class JsonOnlyHttp(Http):
    def get(
        self,
        url: str,
        headers: Mapping[str, str | bytes | None] | None = None,
    ) -> requests.Response:
        raise NotImplementedError

    def get_json(
        self,
        url: str,
        headers: Mapping[str, str | bytes | None] | None = None,
    ) -> JsonOnlyResponse:
        return JsonOnlyResponse(super().get(url, headers))

    def post(
        self,
        url: str,
        json: dict | None = None,
        data: dict | str | list | bytes | None = None,
        headers: Mapping[str, str | bytes | None] | None = None,
    ) -> requests.Response:
        raise NotImplementedError

    def put(
        self,
        url: str,
        json: dict | None = None,
        data: dict | str | list | bytes | None = None,
        headers: Mapping[str, str | bytes | None] | None = None,
    ) -> requests.Response:
        raise NotImplementedError

    def patch(
        self,
        url: str,
        json: dict | None = None,
        data: dict | str | list | bytes | None = None,
        headers: Mapping[str, str | bytes | None] | None = None,
    ) -> requests.Response:
        raise NotImplementedError


class OntologiesHttp(JsonOnlyHttp):
    """
    An HTTP client for the ontologies service.
    """

    def __init__(self) -> None:
        super().__init__(base_url="https://ontologies.canvasmedical.com")

        self._session.headers.update({"Authorization": os.getenv("PRE_SHARED_KEY", "")})


class ScienceHttp(JsonOnlyHttp):
    """
    An HTTP client for the ontologies service.
    """

    def __init__(self) -> None:
        super().__init__(base_url="https://science.canvasmedical.com")

        self._session.headers.update({"Authorization": os.getenv("PRE_SHARED_KEY", "")})


ontologies_http = OntologiesHttp()
science_http = ScienceHttp()

__all__ = __exports__ = (
    "ThreadPoolExecutor",
    "Http",
    "ontologies_http",
    "science_http",
    "batch_get",
    "batch_post",
    "batch_put",
    "batch_patch",
)



from django.contrib.postgres.fields import ArrayField
from django.db import models

from canvas_sdk.v1.data.common import ColorEnum, Origin


class TaskType(models.TextChoices):
    """Choices for task types."""

    TASK = "Task", "Task"
    REMINDER = "Reminder", "Reminder"


class EventType(models.TextChoices):
    """Choices for event types."""

    EVENT_CHART_OPEN = "Chart Open", "Chart Open"


class TaskStatus(models.TextChoices):
    """Choices for task statuses."""

    COMPLETED = "COMPLETED", "Completed"
    CLOSED = "CLOSED", "Closed"
    OPEN = "OPEN", "Open"


class TaskLabelModule(models.TextChoices):
    """Choices for task label modules."""

    CLAIMS = "claims", "Claims"
    TASKS = "tasks", "Tasks"


class Task(models.Model):
    """Task."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_task_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    creator = models.ForeignKey(
        "v1.Staff", on_delete=models.DO_NOTHING, related_name="creator_tasks", null=True
    )
    assignee = models.ForeignKey(
        "v1.Staff", on_delete=models.DO_NOTHING, related_name="assignee_tasks", null=True
    )
    # TODO - uncomment when Team model is created
    # team = models.ForeignKey('v1.Team', on_delete=models.DO_NOTHING, related_name="tasks", null=True)
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, blank=True, related_name="tasks", null=True
    )
    task_type = models.CharField(choices=TaskType.choices)
    tag = models.CharField()
    title = models.CharField()
    due = models.DateTimeField(null=True)
    due_event = models.CharField(choices=EventType.choices, blank=True)
    status = models.CharField(choices=TaskStatus.choices)


class TaskComment(models.Model):
    """TaskComment."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_taskcomment_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    creator = models.ForeignKey(
        "v1.Staff", on_delete=models.DO_NOTHING, related_name="comments", null=True
    )
    task = models.ForeignKey(Task, on_delete=models.DO_NOTHING, related_name="comments", null=True)
    body = models.TextField()


class TaskLabel(models.Model):
    """TaskLabel."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_tasklabel_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    tasks = models.ManyToManyField(Task, related_name="labels", through="TaskTaskLabel")  # type: ignore[var-annotated]
    position = models.IntegerField()
    color = models.CharField(choices=ColorEnum.choices)
    task_association = ArrayField(models.CharField(choices=Origin.choices))
    name = models.CharField()
    active = models.BooleanField()
    modules = ArrayField(models.CharField(choices=TaskLabelModule.choices))


class TaskTaskLabel(models.Model):
    """M2M for Task -> TaskLabels."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_tasktasklabel_001"

    dbid = models.BigIntegerField(primary_key=True)
    task_label = models.ForeignKey(TaskLabel, on_delete=models.DO_NOTHING, null=True)
    task = models.ForeignKey(Task, on_delete=models.DO_NOTHING, null=True)


__exports__ = (
    "TaskType",
    "EventType",
    "TaskStatus",
    "TaskLabelModule",
    "Task",
    "TaskComment",
    "TaskLabel",
    "TaskTaskLabel",
)



from django.core.cache import InvalidCacheBackendError
from django.core.cache import caches as django_caches

from canvas_sdk.caching.base import Cache
from canvas_sdk.caching.exceptions import CacheConfigurationError

caches: dict[tuple[str, str], Cache] = {}


def get_cache(
    driver: str = "default", prefix: str = "", max_timeout_seconds: int | None = None
) -> Cache:
    """Get the cache client based on the specified driver."""
    try:
        key = (driver, prefix)
        connection = django_caches[driver]
        if key not in caches:
            caches[key] = Cache(connection, prefix, max_timeout_seconds)
        return caches[key]
    except InvalidCacheBackendError as error:
        raise CacheConfigurationError(driver) from error


__exports__ = ()



from datetime import date

from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand as BaseCommand


class PastSurgicalHistoryCommand(BaseCommand):
    """A class for managing a Past Surgical History command within a specific note."""

    class Meta:
        key = "surgicalHistory"

    past_surgical_history: str | None = None
    approximate_date: date | None = None
    comment: str | None = Field(max_length=1000, default=None)


__exports__ = ("PastSurgicalHistoryCommand",)



from django.contrib.postgres.fields import ArrayField
from django.db import models
from timezone_utils.fields import TimeZoneField

from canvas_sdk.v1.data.common import (
    AddressState,
    AddressType,
    AddressUse,
    ContactPointState,
    ContactPointSystem,
    ContactPointUse,
    PersonSex,
    TaxIDType,
)


class Staff(models.Model):
    """Staff."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_staff_001"

    def __str__(self) -> str:
        return f"{self.first_name} {self.last_name}"

    id = models.CharField(max_length=32, db_column="key")
    dbid = models.BigIntegerField(db_column="dbid", primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    prefix = models.CharField()
    suffix = models.CharField()
    first_name = models.CharField()
    middle_name = models.CharField()
    last_name = models.CharField()
    maiden_name = models.CharField()
    nickname = models.CharField()
    previous_names = models.JSONField()
    birth_date = models.DateField(null=True)
    sex_at_birth = models.CharField(choices=PersonSex.choices)
    sexual_orientation_term = models.CharField()
    sexual_orientation_code = models.CharField()
    gender_identity_term = models.CharField()
    gender_identity_code = models.CharField()
    preferred_pronouns = models.CharField()
    biological_race_codes = ArrayField(models.CharField())
    biological_race_terms = ArrayField(models.CharField())
    cultural_ethnicity_codes = ArrayField(models.CharField())
    cultural_ethnicity_terms = ArrayField(models.CharField())
    last_known_timezone = TimeZoneField(null=True)
    active = models.BooleanField()
    primary_practice_location = models.ForeignKey(
        "v1.PracticeLocation", on_delete=models.DO_NOTHING, null=True
    )
    npi_number = models.CharField()
    nadean_number = models.CharField()
    group_npi_number = models.CharField()
    bill_through_organization = models.BooleanField()
    tax_id = models.CharField()
    tax_id_type = models.CharField(choices=TaxIDType.choices)
    spi_number = models.CharField()
    # TODO - uncomment when Language is developed
    # language = models.ForeignKey('v1.Language', on_delete=models.DO_NOTHING, related_name="staff_speakers", null=True)
    # language_secondary = models.ForeignKey('v1.Language', on_delete=models.DO_NOTHING, related_name="staff_secondary_speakers", null=True)
    personal_meeting_room_link = models.URLField(null=True)
    state = models.JSONField()
    user = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    schedule_column_ordering = models.IntegerField()
    default_supervising_provider = models.ForeignKey(
        "v1.Staff", on_delete=models.DO_NOTHING, related_name="supervising_team", null=True
    )


class StaffContactPoint(models.Model):
    """StaffContactPoint."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_staffcontactpoint_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    system = models.CharField(choices=ContactPointSystem.choices)
    value = models.CharField()
    use = models.CharField(choices=ContactPointUse.choices)
    use_notes = models.CharField()
    rank = models.IntegerField()
    state = models.CharField(choices=ContactPointState.choices)
    staff = models.ForeignKey(Staff, on_delete=models.DO_NOTHING, related_name="telecom")


class StaffAddress(models.Model):
    """StaffAddress."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_staffaddress_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    line1 = models.CharField()
    line2 = models.CharField()
    city = models.CharField()
    district = models.CharField()
    state_code = models.CharField()
    postal_code = models.CharField()
    use = models.CharField(choices=AddressUse.choices)
    type = models.CharField(choices=AddressType.choices)
    longitude = models.FloatField()
    latitude = models.FloatField()
    start = models.DateField()
    end = models.DateField()
    country = models.CharField()
    state = models.CharField(choices=AddressState.choices)
    staff = models.ForeignKey(Staff, on_delete=models.DO_NOTHING, related_name="addresses")


class StaffPhoto(models.Model):
    """StaffPhoto."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_staffphoto_001"

    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    staff = models.ForeignKey(Staff, on_delete=models.CASCADE, related_name="photos")
    url = models.CharField(
        default="https://d3hn0m4rbsz438.cloudfront.net/avatar1.png", max_length=512
    )
    title = models.CharField(max_length=255, blank=True, default="")


__exports__ = ("Staff", "StaffContactPoint", "StaffAddress", "StaffPhoto")



from typing import Any, Self

from pydantic import model_validator
from pydantic_core import InitErrorDetails

from canvas_sdk.effects.base import EffectType, _BaseEffect
from canvas_sdk.v1.data import CanvasUser


class UpdateUserEffect(_BaseEffect):
    """
    An Effect that will update the user properties.
    """

    class Meta:
        effect_type = EffectType.UPDATE_USER

    # A set to track which fields have been modified.
    _dirty_keys: set[str] = set()

    def __init__(self, /, **data: Any) -> None:
        super().__init__(**data)
        self._dirty_keys = set()
        self._dirty_keys.update(data.keys())

    def __setattr__(self, name: str, value: Any) -> None:
        """Set an attribute and mark it as dirty unless excluded."""
        if not name.startswith("_"):
            self._dirty_keys.add(name)
        super().__setattr__(name, value)

    def is_dirty(self, key: str) -> bool:
        """Returns True if the given property has been modified (i.e. marked as dirty), False otherwise."""
        return key in self._dirty_keys

    user_dbid: int
    email: str | None = None
    phone_number: str | None = None

    @property
    def values(self) -> dict[str, Any]:
        """The user's values."""
        result = {}

        for key in self._dirty_keys:
            result[key] = getattr(self, key)

        result["dbid"] = self.user_dbid
        return result

    @property
    def effect_payload(self) -> dict[str, Any]:
        """The payload of the effect."""
        return {"data": self.values}

    @model_validator(mode="after")
    def check_exclusive_fields(self) -> Self:
        """Check that at least one of mutually exclusive field is set."""
        if self.email is None and self.phone_number is None:
            raise ValueError("one of 'email', 'phone_number' is required")

        return self

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        user_exists = CanvasUser.objects.filter(dbid=self.user_dbid).exists()

        if not user_exists:
            errors.append(
                self._create_error_detail(
                    "value",
                    "User does not exist",
                    self.user_dbid,
                )
            )

        return errors


__exports__ = ("UpdateUserEffect",)



from django.contrib.postgres.fields import ArrayField
from django.db import models


class NoteTypeCategories(models.TextChoices):
    """Note type categories."""

    MESSAGE = "message", "Message"
    LETTER = "letter", "Letter"
    INPATIENT = "inpatient", "Inpatient Visit Note"
    REVIEW = "review", "Chart Review Note"
    ENCOUNTER = "encounter", "Encounter Note"
    APPOINTMENT = "appointment", "Appointment Note"
    TASK = "task", "Task"
    DATA = "data", "Data"
    CCDA = "ccda", "C-CDA"
    SCHEDULE_EVENT = "schedule_event", "Schedule Event"


class PracticeLocationPOS(models.TextChoices):
    """Practice Location POS."""

    PHARMACY = "01", "Pharmacy"
    TELEHEALTH = "02", "Telehealth"
    SCHOOL = "03", "Education Facility"
    HOMELESS_SHELTER = "04", "Homeless Shelter"
    PRISON = "09", "Prison"
    TELEHEALTH_IN_PATIENT_HOME = "10", "Telehealth in Patient's Home"
    OFFICE = "11", "Office"
    HOME = "12", "Home"
    ASSISTED_LIVING = "13", "Asssisted Living Facility"
    GROUP_HOME = "14", "Group Home"
    MOBILE = "15", "Mobile Unit"
    WALK_IN_RETAIL = "17", "Walk-In Retail Health Clinic"
    OFF_CAMPUS_OUTPATIENT_HOSPITAL = "19", "Off-Campus Outpatient Hospital"
    URGENT_CARE = "20", "Urgent Care Facility"
    INPATIENT_HOSPITAL = "21", "Inpatient Hospital"
    ON_CAMPUS_OUTPATIENT_HOSPITAL = "22", "On-Campus Outpatient Hospital"
    ER_HOSPITAL = "23", "Emergency Room Hospital"
    AMBULATORY_SURGERY_CENTER = "24", "Ambulatory Surgery Center"
    BIRTHING_CENTER = "25", "Birthing Center"
    MILITARY_FACILITY = "26", "Military Treatment Facility"
    STREET = "27", "Outreach Site / Street"
    SNF = "31", "Skilled Nursing Facility"
    NURSING = "32", "Nursing Facility"
    CUSTODIAL = "33", "Custodial Care Facility"
    HOSPICE = "34", "Hospice"
    AMBULANCE_LAND = "41", "Ambulance Land"
    AMBULANCE_AIR_WATER = "42", "Ambulance Air or Water"
    INDEPENDENT_CLINIC = "49", "Independent Clinic"
    FQHC = "50", "Federally Qualified Health Center"
    PSYCH = "51", "Inpatient Psychiatric Facility"
    PSYCH_PARTIAL = "52", "Inpatient Psychiatric Facility - Partial Hospitalization"
    MENTAL_HEALTH_CENTER = "53", "Community Mental Health Center"
    INTERMEDIATE_MENTAL = "54", "Intermediate Care Facility for Mentally Retarded"
    SUBSTANCE_RESIDENTIAL = "55", "Residential Substance Abuse Treatment Facility"
    PSYCH_RESIDENTIAL = "56", "Psychiatric Residential Treatment Center"
    SUBSTANCE_NON_RESIDENTIAL = "57", "Non-Residential Substance Abuse Treatment Facility"
    MASS_IMMUNIZATION = "60", "Mass Immunization Center"
    INPATIENT_REHAB = "61", "Inpatient Rehabilitation Facility"
    OUTPATIENT_REHAB = "62", "Outpatient Rehabilitation Facility"
    ESRD = "65", "End-Stage Renal Disease Treatment Facility"
    PUBLIC_CLINIC = "71", "State or Local Public Health Clinic"
    RURAL_CLINIC = "72", "Rural Health Clinic"
    INDEPENDENT_LAB = "81", "Independent Laboratory"
    OTHER = "99", "Other Place of Service"


class NoteTypes(models.TextChoices):
    """Note types."""

    MESSAGE = "message", "Message"
    LETTER = "letter", "Letter"
    INPATIENT = "inpatient", "Inpatient Visit Note"
    REVIEW = "review", "Chart Review Note"
    VOICE = "voice", "Phone Call Note"
    VIDEO = "video", "Video Call Note"
    OFFICE = "office", "Office Visit Note"
    LAB = "lab", "Lab Visit Note"
    HOME = "home", "Home Visit Note"
    GROUP = "group", "Group Visit Note"
    APPOINTMENT = "appointment", "Appointment Note"
    OFFSITE = "offsite", "Other Offsite Visit Note"
    SEARCH = "search", "Search"
    TASK = "task", "Task"
    DATA = "data", "Data"
    CCDA = "ccda", "C-CDA Import"


class NoteType(models.Model):
    """NoteType."""

    objects: models.Manager["NoteType"]

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_notetype_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    system = models.CharField()
    version = models.CharField()
    code = models.CharField()
    display = models.CharField()
    user_selected = models.BooleanField()
    name = models.CharField()
    icon = models.CharField()
    category = models.CharField(choices=NoteTypeCategories.choices)
    rank = models.PositiveIntegerField()
    is_default_appointment_type = models.BooleanField()
    is_scheduleable = models.BooleanField()
    is_telehealth = models.BooleanField()
    is_billable = models.BooleanField()
    defer_place_of_service_to_practice_location = models.BooleanField()
    available_places_of_service = ArrayField(models.CharField(choices=PracticeLocationPOS.choices))
    default_place_of_service = models.CharField(choices=PracticeLocationPOS.choices)
    is_system_managed = models.BooleanField()
    is_visible = models.BooleanField()
    is_active = models.BooleanField()
    unique_identifier = models.UUIDField()
    deprecated_at = models.DateTimeField()
    is_patient_required = models.BooleanField()
    allow_custom_title = models.BooleanField()
    is_scheduleable_via_patient_portal = models.BooleanField()
    online_duration = models.IntegerField()


class Note(models.Model):
    """Note."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_note_001"

    id = models.CharField(max_length=32)
    dbid = models.BigIntegerField(db_column="dbid", primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="notes", null=True
    )
    provider = models.ForeignKey(
        "v1.Staff", on_delete=models.DO_NOTHING, related_name="notes", null=True
    )
    note_type = models.CharField(choices=NoteTypes.choices, null=True)
    note_type_version = models.ForeignKey(
        "v1.NoteType", on_delete=models.DO_NOTHING, related_name="notes", null=True
    )
    title = models.TextField()
    body = models.JSONField()
    originator = models.ForeignKey("v1.CanvasUser", on_delete=models.DO_NOTHING, null=True)
    last_modified_by_staff = models.ForeignKey("v1.Staff", on_delete=models.DO_NOTHING, null=True)
    checksum = models.CharField()
    billing_note = models.TextField()
    # TODO -implement InpatientStay model
    # inpatient_stay = models.ForeignKey("v1.InpatientStay", on_delete=models.DO_NOTHING, null=True)
    related_data = models.JSONField()
    location = models.ForeignKey("v1.PracticeLocation", on_delete=models.DO_NOTHING, null=True)
    datetime_of_service = models.DateTimeField()
    place_of_service = models.CharField()


__exports__ = (
    "NoteTypeCategories",
    "PracticeLocationPOS",
    "NoteTypes",
    "NoteType",
    "Note",
)



from typing import Any
from uuid import UUID

from pydantic_core import InitErrorDetails

from canvas_sdk.effects.note.base import AppointmentABC
from canvas_sdk.v1.data import NoteType, Patient
from canvas_sdk.v1.data.note import NoteTypeCategories


class ScheduleEvent(AppointmentABC):
    """
    Effect to create a schedule event.

    Attributes:
        note_type_id (UUID | str): The ID of the note type.
        patient_id (str | None): The ID of the patient, if applicable.
        description (str | None): The description of the schedule event, if applicable.
    """

    class Meta:
        effect_type = "SCHEDULE_EVENT"

    note_type_id: UUID | str
    patient_id: str | None = None
    description: str | None = None

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        """
        Validates the schedule event note type and returns a list of error details if validation fails.

        Args:
            method (Any): The method being validated.

        Returns:
            list[InitErrorDetails]: A list of error details for invalid schedule event note types.
        """
        errors = super()._get_error_details(method)

        note_type = (
            NoteType.objects.values("category", "is_patient_required", "allow_custom_title")
            .filter(id=self.note_type_id)
            .first()
        )

        if not note_type:
            errors.append(
                self._create_error_detail(
                    "value",
                    f"Note type with ID {self.note_type_id} does not exist.",
                    self.note_type_id,
                )
            )
            return errors

        if note_type["category"] != NoteTypeCategories.SCHEDULE_EVENT:
            errors.append(
                self._create_error_detail(
                    "value",
                    f"Schedule event note type must be of type: {NoteTypeCategories.SCHEDULE_EVENT}.",
                    self.note_type_id,
                )
            )

        if note_type["is_patient_required"] and not self.patient_id:
            errors.append(
                self._create_error_detail(
                    "value",
                    "Patient ID is required for this note type.",
                    self.note_type_id,
                )
            )

        if not note_type["allow_custom_title"] and self.description:
            errors.append(
                self._create_error_detail(
                    "value",
                    "Description is not allowed for this note type.",
                    self.note_type_id,
                )
            )

        return errors


class Appointment(AppointmentABC):
    """
    Effect to create an appointment.

    Attributes:
        appointment_note_type_id (UUID | str): The ID of the appointment note type.
        meeting_link (str | None): The meeting link for the appointment, if any.
        patient_id (str): The ID of the patient.
    """

    class Meta:
        effect_type = "APPOINTMENT"

    appointment_note_type_id: UUID | str
    meeting_link: str | None = None
    patient_id: str

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        """
        Validates the appointment note type and returns a list of error details if validation fails.

        Args:
            method (Any): The method being validated.

        Returns:
            list[InitErrorDetails]: A list of error details for invalid appointment note types or missing patient IDs.
        """
        errors = super()._get_error_details(method)

        if not Patient.objects.filter(id=self.patient_id).exists():
            errors.append(
                self._create_error_detail(
                    "value",
                    "Patient with ID {self.patient_id} does not exist.",
                    self.patient_id,
                )
            )

        category, is_scheduleable = NoteType.objects.values_list("category", "is_scheduleable").get(
            id=self.appointment_note_type_id
        )
        if category != NoteTypeCategories.ENCOUNTER:
            errors.append(
                self._create_error_detail(
                    "value",
                    f"Appointment note type must be of type, {NoteTypeCategories.ENCOUNTER} but got: {category}.",
                    self.appointment_note_type_id,
                )
            )

        if not is_scheduleable:
            errors.append(
                self._create_error_detail(
                    "value",
                    "Appointment note type must be scheduleable.",
                    self.appointment_note_type_id,
                )
            )

        return errors


__exports__ = ("ScheduleEvent", "Appointment")



from django.db import models

from canvas_sdk.v1.data.common import TaxIDType


class Organization(models.Model):
    """Organization."""

    class Meta:
        managed = False
        db_table = "canvas_sdk_data_api_organization_001"

    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    full_name = models.CharField()
    short_name = models.CharField()
    subdomain = models.CharField()
    logo_url = models.CharField()
    background_image_url = models.CharField()
    background_gradient = models.CharField()
    active = models.BooleanField()
    tax_id = models.CharField(null=True)
    tax_id_type = models.CharField(choices=TaxIDType.choices)
    group_npi_number = models.CharField()
    group_taxonomy_number = models.CharField()
    include_zz_qualifier = models.BooleanField()
    main_location = models.OneToOneField("v1.PracticeLocation", on_delete=models.DO_NOTHING)


__exports__ = ("Organization",)



from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand as BaseCommand


class PerformCommand(BaseCommand):
    """A class for managing a Perform command within a specific note."""

    class Meta:
        key = "perform"

    cpt_code: str | None = Field(default=None, json_schema_extra={"commands_api_name": "perform"})
    notes: str | None = None


__exports__ = ("PerformCommand",)
